@@ -1,7 +1,7 @@
 using System;
 using System.IO;
 using System.Text;
-using System.Threading;
+using ST = System.Threading;
 using System.Configuration;
 using System.Globalization;
 using System.Reflection;
@@ -12,36 +12,32 @@ using System.Runtime.InteropServices;
 
 using Mono.Debugger.Languages;
 
-namespace Mono.Debugger.Backends
+namespace Mono.Debugger.Backend
 {
-	internal delegate void ChildOutputHandler (string output);
+	internal delegate void ChildOutputHandler (bool is_stderr, string output);
 
-	internal class Inferior : TargetMemoryAccess, ITargetNotification, IDisposable
+	internal class Inferior : TargetAccess, ITargetNotification, IDisposable
 	{
 		protected IntPtr server_handle;
+		protected IntPtr io_data;
 		protected Bfd bfd;
-		protected BfdDisassembler bfd_disassembler;
 		protected ThreadManager thread_manager;
 
 		protected readonly ProcessStart start;
 
 		protected readonly BfdContainer bfd_container;
-		protected readonly SymbolTableCollection symtab_collection;
 		protected readonly ProcessServant process;
 		protected readonly DebuggerErrorHandler error_handler;
 		protected readonly BreakpointManager breakpoint_manager;
 		protected readonly AddressDomain address_domain;
 		protected readonly bool native;
 
-		protected ChildOutputHandler stdout_handler;
-		protected ChildOutputHandler stderr_handler;
-
 		int child_pid;
 		bool initialized;
 		bool has_target;
 		bool pushed_regs;
 
-		TargetInfo target_info;
+		TargetMemoryInfo target_info;
 		Architecture arch;
 
 		bool has_signals;
@@ -67,7 +63,9 @@ namespace Mono.Debugger.Backends
 		static extern TargetError mono_debugger_server_initialize_thread (IntPtr handle, int child_pid);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_spawn (IntPtr handle, string working_directory, string[] argv, string[] envp, out int child_pid, ChildOutputHandler stdout_handler, ChildOutputHandler stderr_handler, out IntPtr error);
+		static extern TargetError mono_debugger_server_io_thread_main (IntPtr io_data, ChildOutputHandler output_handler);
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_spawn (IntPtr handle, string working_directory, string[] argv, string[] envp, out int child_pid, out IntPtr io_data, out IntPtr error);
 
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_attach (IntPtr handle, int child_pid);
@@ -84,6 +82,9 @@ namespace Mono.Debugger.Backends
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_continue (IntPtr handle);
 
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_resume (IntPtr handle);
+
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_detach (IntPtr handle);
 
@@ -103,17 +104,26 @@ namespace Mono.Debugger.Backends
 		static extern TargetError mono_debugger_server_call_method (IntPtr handle, long method_address, long method_argument1, long method_argument2, long callback_argument);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_call_method_1 (IntPtr handle, long method_address, long method_argument, string string_argument, long callback_argument);
+		static extern TargetError mono_debugger_server_call_method_1 (IntPtr handle, long method_address, long method_argument, long data_argument, long data_argument2, string string_argument, long callback_argument);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_call_method_2 (IntPtr handle, long method_address, long method_argument, long callback_argument);
+		static extern TargetError mono_debugger_server_call_method_2 (IntPtr handle, long method_address, int data_size, IntPtr data, long callback_argument);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_abort_invoke (IntPtr handle);
+		static extern TargetError mono_debugger_server_call_method_3 (IntPtr handle, long method_address, long method_argument, long address_argument, int blob_size, IntPtr blob_data, long callback_argument);
+
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_mark_rti_frame (IntPtr handle);
+
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_abort_invoke (IntPtr handle, long stack_pointer);
 
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_call_method_invoke (IntPtr handle, long invoke_method, long method_address, int num_params, int blob_size, IntPtr param_data, IntPtr offset_data, IntPtr blob_data, long callback_argument, bool debug);
 
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_execute_instruction (IntPtr handle, IntPtr instruction, int insn_size, bool update_ip);
+
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_insert_breakpoint (IntPtr handle, long address, out int breakpoint);
 
@@ -151,7 +161,7 @@ namespace Mono.Debugger.Backends
 		static extern IntPtr mono_debugger_server_create_inferior (IntPtr breakpoint_manager);
 
 		[DllImport("monodebuggerserver")]
-		static extern ChildEventType mono_debugger_server_dispatch_event (IntPtr handle, int status, out long arg, out long data1, out long data2);
+		static extern ChildEventType mono_debugger_server_dispatch_event (IntPtr handle, int status, out long arg, out long data1, out long data2, out int opt_data_size, out IntPtr opt_data);
 
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_get_signal_info (IntPtr handle, out IntPtr data);
@@ -163,7 +173,7 @@ namespace Mono.Debugger.Backends
 		static extern TargetError mono_debugger_server_get_application (IntPtr handle, out string exe_file, out string cwd, out int nargs, out IntPtr data);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_init_after_fork (IntPtr handle);
+		static extern TargetError mono_debugger_server_detach_after_fork (IntPtr handle);
 
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_push_registers (IntPtr handle, out long new_rsp);
@@ -172,7 +182,10 @@ namespace Mono.Debugger.Backends
 		static extern TargetError mono_debugger_server_pop_registers (IntPtr handle);
 
 		[DllImport("monodebuggerserver")]
-		static extern void mono_debugger_server_set_notification (IntPtr handle, long address);
+		static extern TargetError mono_debugger_server_get_callback_frame (IntPtr handle, long stack_pointer, bool exact_match, IntPtr registers);
+
+		[DllImport("monodebuggerserver")]
+		static extern void mono_debugger_server_set_runtime_info (IntPtr handle, IntPtr mono_runtime_info);
 
 		internal enum ChildEventType {
 			NONE = 0,
@@ -213,8 +226,9 @@ namespace Mono.Debugger.Backends
 			public readonly long Data1;
 			public readonly long Data2;
 
-			public ChildEvent (ChildEventType type, long arg,
-					   long data1, long data2)
+			public readonly byte[] CallbackData;
+
+			public ChildEvent (ChildEventType type, long arg, long data1, long data2)
 			{
 				this.Type = type;
 				this.Argument = arg;
@@ -222,6 +236,13 @@ namespace Mono.Debugger.Backends
 				this.Data2 = data2;
 			}
 
+			public ChildEvent (ChildEventType type, long arg, long data1, long data2,
+					   byte[] callback_data)
+				: this (type, arg, data1, data2)
+			{
+				this.CallbackData = callback_data;
+			}
+
 			public override string ToString ()
 			{
 				return String.Format ("ChildEvent ({0}:{1}:{2:x}:{3:x})",
@@ -237,12 +258,20 @@ namespace Mono.Debugger.Backends
 			public int SIGINT;
 			public int SIGCHLD;
 
+			public int SIGFPE;
+			public int SIGQUIT;
+			public int SIGABRT;
+			public int SIGSEGV;
+			public int SIGILL;
+			public int SIGBUS;
+
+			public int Kernel_SIGRTMIN;
 			public int MonoThreadAbortSignal;
 
 			public override string ToString ()
 			{
-				return String.Format ("SignalInfo ({0}:{1}:{2}:{3} - {4})",
-						      SIGKILL, SIGSTOP, SIGINT, SIGCHLD,
+				return String.Format ("SignalInfo ({0}:{1}:{2}:{3}:{4} - {5})",
+						      SIGKILL, SIGSTOP, SIGINT, SIGCHLD, Kernel_SIGRTMIN,
 						      MonoThreadAbortSignal);
 			}
 		}
@@ -287,7 +316,6 @@ namespace Mono.Debugger.Backends
 			inferior.bfd = bfd;
 
 			inferior.arch = inferior.bfd.Architecture;
-			inferior.bfd_disassembler = inferior.bfd.GetDisassembler (inferior);
 
 			return inferior;
 		}
@@ -319,39 +347,82 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void CallMethod (TargetAddress method, long arg1, string arg2,
-					long callback_arg)
+		public void CallMethod (TargetAddress method, long arg1, long arg2, long arg3,
+					string arg4, long callback_arg)
 		{
 			check_disposed ();
 
 			TargetState old_state = change_target_state (TargetState.Running);
 			try {
 				check_error (mono_debugger_server_call_method_1 (
-					server_handle, method.Address, arg1,
-					arg2, callback_arg));
+					server_handle, method.Address, arg1, arg2, arg3,
+					arg4, callback_arg));
 			} catch {
 				change_target_state (old_state);
 				throw;
 			}
 		}
 
-		public void CallMethod (TargetAddress method, long arg1, long callback_arg)
+		public void CallMethod (TargetAddress method, byte[] data, long callback_arg)
 		{
 			check_disposed ();
 
 			TargetState old_state = change_target_state (TargetState.Running);
+
+			IntPtr data_ptr = IntPtr.Zero;
+			int data_size = data != null ? data.Length : 0;
+
 			try {
+				if (data != null) {
+					data_ptr = Marshal.AllocHGlobal (data_size);
+					Marshal.Copy (data, 0, data_ptr, data_size);
+				}
+
 				check_error (mono_debugger_server_call_method_2 (
-					server_handle, method.Address, arg1, callback_arg));
+					server_handle, method.Address,
+					data_size, data_ptr, callback_arg));
 			} catch {
 				change_target_state (old_state);
 				throw;
+			} finally {
+				if (data_ptr != IntPtr.Zero)
+					Marshal.FreeHGlobal (data_ptr);
 			}
 		}
 
-		public void RuntimeInvoke (Thread target, TargetAddress invoke_method,
+		public void CallMethod (TargetAddress method, long method_argument,
+					TargetObject obj, long callback_arg)
+		{
+			check_disposed ();
+
+			byte[] blob = null;
+			long address = 0;
+
+			if (obj.Location.HasAddress)
+				address = obj.Location.GetAddress (this).Address;
+			else
+				blob = obj.Location.ReadBuffer (this, obj.Type.Size);
+
+			IntPtr blob_data = IntPtr.Zero;
+			try {
+				if (blob != null) {
+					blob_data = Marshal.AllocHGlobal (blob.Length);
+					Marshal.Copy (blob, 0, blob_data, blob.Length);
+				}
+
+				check_error (mono_debugger_server_call_method_3 (
+					server_handle, method.Address, method_argument,
+					address, blob != null ? blob.Length : 0, blob_data, callback_arg));
+			} finally {
+				if (blob_data != IntPtr.Zero)
+					Marshal.FreeHGlobal (blob_data);
+			}
+		}
+
+		public void RuntimeInvoke (TargetAddress invoke_method,
 					   TargetAddress method_argument,
-					   TargetObject object_argument, TargetObject[] param_objects,
+					   TargetObject object_argument,
+					   TargetObject[] param_objects,
 					   long callback_arg, bool debug)
 		{
 			check_disposed ();
@@ -374,10 +445,10 @@ namespace Mono.Debugger.Backends
 					continue;
 				if (obj.Location.HasAddress) {
 					blob_offsets [i] = -1;
-					addresses [i] = obj.Location.Address.Address;
+					addresses [i] = obj.Location.GetAddress (this).Address;
 					continue;
 				}
-				blobs [i] = obj.Location.ReadBuffer (target, obj.Type.Size);
+				blobs [i] = obj.Location.ReadBuffer (this, obj.Type.Size);
 				blob_offsets [i] = blob_size;
 				blob_size += blobs [i].Length;
 			}
@@ -417,9 +488,35 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void AbortInvoke ()
+		public void ExecuteInstruction (byte[] instruction, bool update_ip)
 		{
-			check_error (mono_debugger_server_abort_invoke (server_handle));
+			check_disposed ();
+
+			IntPtr data = IntPtr.Zero;
+			try {
+				data = Marshal.AllocHGlobal (instruction.Length);
+				Marshal.Copy (instruction, 0, data, instruction.Length);
+
+				check_error (mono_debugger_server_execute_instruction (
+					server_handle, data, instruction.Length, update_ip));
+			} finally {
+				Marshal.FreeHGlobal (data);
+			}
+		}
+
+		public void MarkRuntimeInvokeFrame ()
+		{
+			check_error (mono_debugger_server_mark_rti_frame (server_handle));
+		}
+
+		public bool AbortInvoke (TargetAddress stack_pointer)
+		{
+			TargetError result = mono_debugger_server_abort_invoke (
+				server_handle, stack_pointer.Address);
+			if (result == TargetError.NoCallbackFrame)
+				return false;
+			check_error (result);
+			return true;
 		}
 
 		public int InsertBreakpoint (TargetAddress address)
@@ -495,7 +592,12 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public int Run (bool redirect_fds)
+		void io_thread_main ()
+		{
+			mono_debugger_server_io_thread_main (io_data, process.OnTargetOutput);
+		}
+
+		public int Run ()
 		{
 			if (has_target)
 				throw new TargetException (TargetError.AlreadyHaveTarget);
@@ -504,15 +606,9 @@ namespace Mono.Debugger.Backends
 
 			IntPtr error;
 
-			stdout_handler = new ChildOutputHandler (inferior_stdout_handler);
-			stderr_handler = new ChildOutputHandler (inferior_stderr_handler);
-
 			TargetError result = mono_debugger_server_spawn (
 				server_handle, start.WorkingDirectory, start.CommandLineArguments,
-				start.Environment, out child_pid,
-				stdout_handler,
-				stderr_handler,
-				out error);
+				start.Environment, out child_pid, out io_data, out error);
 			if (result != TargetError.None) {
 				string message = Marshal.PtrToStringAuto (error);
 				g_free (error);
@@ -521,6 +617,10 @@ namespace Mono.Debugger.Backends
 					TargetError.CannotStartTarget, message);
 			}
 
+			ST.Thread io_thread = new ST.Thread (new ST.ThreadStart (io_thread_main));
+			io_thread.IsBackground = true;
+			io_thread.Start ();
+
 			return child_pid;
 		}
 
@@ -590,8 +690,12 @@ namespace Mono.Debugger.Backends
 			long arg, data1, data2;
 			ChildEventType message;
 
+			int opt_data_size;
+			IntPtr opt_data;
+
 			message = mono_debugger_server_dispatch_event (
-				server_handle, status, out arg, out data1, out data2);
+				server_handle, status, out arg, out data1, out data2,
+				out opt_data_size, out opt_data);
 
 			switch (message) {
 			case ChildEventType.CHILD_EXITED:
@@ -612,10 +716,18 @@ namespace Mono.Debugger.Backends
 				break;
 			}
 
+			if (opt_data_size > 0) {
+				byte[] data = new byte [opt_data_size];
+				Marshal.Copy (opt_data, data, 0, opt_data_size);
+				g_free (opt_data);
+
+				return new ChildEvent (message, arg, data1, data2, data);
+			}
+
 			return new ChildEvent (message, arg, data1, data2);
 		}
 
-		public static TargetInfo GetTargetInfo (AddressDomain domain)
+		public static TargetInfo GetTargetInfo ()
 		{
 			int target_int_size, target_long_size, target_addr_size, is_bigendian;
 			check_error (mono_debugger_server_get_target_info
@@ -623,7 +735,18 @@ namespace Mono.Debugger.Backends
 				 out target_addr_size, out is_bigendian));
 
 			return new TargetInfo (target_int_size, target_long_size,
-					       target_addr_size, is_bigendian != 0, domain);
+					       target_addr_size, is_bigendian != 0);
+		}
+
+		public static TargetMemoryInfo GetTargetMemoryInfo (AddressDomain domain)
+		{
+			int target_int_size, target_long_size, target_addr_size, is_bigendian;
+			check_error (mono_debugger_server_get_target_info
+				(out target_int_size, out target_long_size,
+				 out target_addr_size, out is_bigendian));
+
+			return new TargetMemoryInfo (target_int_size, target_long_size,
+						     target_addr_size, is_bigendian != 0, domain);
 		}
 
 		public static string GetFileContents (string filename)
@@ -655,7 +778,7 @@ namespace Mono.Debugger.Backends
 				g_free (data);
 			}
 
-			target_info = GetTargetInfo (address_domain);
+			target_info = GetTargetMemoryInfo (address_domain);
 
 			try {
 				bfd = bfd_container.AddFile (
@@ -674,8 +797,6 @@ namespace Mono.Debugger.Backends
 			bfd_container.SetupInferior (target_info, bfd);
 
 			arch = bfd.Architecture;
-
-			bfd_disassembler = bfd.GetDisassembler (this);
 		}
 
 		public void InitializeModules ()
@@ -707,57 +828,37 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		void inferior_stdout_handler (string line)
-		{
-			if (TargetOutput != null)
-				TargetOutput (false, line);
-		}
-
-		void inferior_stderr_handler (string line)
-		{
-			if (TargetOutput != null)
-				TargetOutput (true, line);
-		}
-
-		//
-		// ITargetInfo
-		//
-
-		public int TargetIntegerSize {
+		public override int TargetIntegerSize {
 			get {
 				return target_info.TargetIntegerSize;
 			}
 		}
 
-		public int TargetLongIntegerSize {
+		public override int TargetLongIntegerSize {
 			get {
 				return target_info.TargetLongIntegerSize;
 			}
 		}
 
-		public int TargetAddressSize {
+		public override int TargetAddressSize {
 			get {
 				return target_info.TargetAddressSize;
 			}
 		}
 
-		public bool IsBigEndian {
+		public override bool IsBigEndian {
 			get {
 				return target_info.IsBigEndian;
 			}
 		}
 
-		//
-		// TargetMemoryAccess
-		//
-
-		public AddressDomain AddressDomain {
+		public override AddressDomain AddressDomain {
 			get {
 				return address_domain;
 			}
 		}
 
-		public override TargetInfo TargetInfo {
+		public override TargetMemoryInfo TargetMemoryInfo {
 			get {
 				return target_info;
 			}
@@ -877,13 +978,13 @@ namespace Mono.Debugger.Backends
 			return new TargetBlob (retval, target_info);
 		}
 
-		public bool CanWrite {
+		public override bool CanWrite {
 			get {
 				return true;
 			}
 		}
 
-		public void WriteBuffer (TargetAddress address, byte[] buffer)
+		public override void WriteBuffer (TargetAddress address, byte[] buffer)
 		{
 			check_disposed ();
 			IntPtr data = IntPtr.Zero;
@@ -900,7 +1001,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void WriteByte (TargetAddress address, byte value)
+		public override void WriteByte (TargetAddress address, byte value)
 		{
 			check_disposed ();
 			IntPtr data = IntPtr.Zero;
@@ -916,7 +1017,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void WriteInteger (TargetAddress address, int value)
+		public override void WriteInteger (TargetAddress address, int value)
 		{
 			check_disposed ();
 			IntPtr data = IntPtr.Zero;
@@ -932,7 +1033,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void WriteLongInteger (TargetAddress address, long value)
+		public override void WriteLongInteger (TargetAddress address, long value)
 		{
 			check_disposed ();
 			IntPtr data = IntPtr.Zero;
@@ -948,7 +1049,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void WriteAddress (TargetAddress address, TargetAddress value)
+		public override void WriteAddress (TargetAddress address, TargetAddress value)
 		{
 			check_disposed ();
 			switch (TargetAddressSize) {
@@ -966,16 +1067,21 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public override int InsertBreakpoint (Breakpoint bpt, TargetAddress address)
+		internal override void InsertBreakpoint (BreakpointHandle handle,
+							 TargetAddress address, int domain)
 		{
-			return breakpoint_manager.InsertBreakpoint (this, bpt, address);
+			breakpoint_manager.InsertBreakpoint (this, handle, address, domain);
+		}
+
+		internal override void RemoveBreakpoint (BreakpointHandle handle)
+		{
+			breakpoint_manager.RemoveBreakpoint (this, handle);
 		}
 
 		//
 		// IInferior
 		//
 
-		public event TargetOutputHandler TargetOutput;
 		public event StateChangedHandler StateChanged;
 
 		TargetState target_state = TargetState.NoTarget;
@@ -1030,6 +1136,19 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
+		public void Resume ()
+		{
+			check_disposed ();
+
+			TargetState old_state = change_target_state (TargetState.Running);
+			try {
+				check_error (mono_debugger_server_resume (server_handle));
+			} catch {
+				change_target_state (old_state);
+				throw;
+			}
+		}
+
 		// <summary>
 		//   Stop the inferior.
 		//   Returns true if it actually stopped the inferior and false if it was
@@ -1112,14 +1231,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public Disassembler Disassembler {
-			get {
-				check_disposed ();
-				return bfd_disassembler;
-			}
-		}
-
-		internal override Architecture Architecture {
+		internal Architecture Architecture {
 			get {
 				check_disposed ();
 				return arch;
@@ -1152,11 +1264,20 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void SetRegisters (Registers registers)
+		public override void SetRegisters (Registers registers)
 		{
 			IntPtr buffer = IntPtr.Zero;
 			try {
 				int count = arch.CountRegisters;
+
+				Registers old_regs = GetRegisters ();
+				for (int i = 0; i < count; i++) {
+					if (registers [i] == null)
+						continue;
+					if (!registers [i].Valid)
+						registers [i].SetValue (old_regs [i].Value);
+				}
+
 				int buffer_size = count * 8;
 				buffer = Marshal.AllocHGlobal (buffer_size);
 				Marshal.Copy (registers.Values, 0, buffer, count);
@@ -1208,10 +1329,10 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public void InitializeAfterFork ()
+		public void DetachAfterFork ()
 		{
-			check_error (mono_debugger_server_init_after_fork (server_handle));
-			breakpoint_manager.InitializeAfterFork (this);
+			mono_debugger_server_detach_after_fork (server_handle);
+			Dispose ();
 		}
 
 		public TargetAddress PushRegisters ()
@@ -1228,9 +1349,31 @@ namespace Mono.Debugger.Backends
 			check_error (mono_debugger_server_pop_registers (server_handle));
 		}
 
-		internal void SetNotificationAddress (TargetAddress notification)
+		internal Registers GetCallbackFrame (TargetAddress stack_pointer, bool exact_match)
 		{
-			mono_debugger_server_set_notification (server_handle, notification.Address);
+			IntPtr buffer = IntPtr.Zero;
+			try {
+				int count = arch.CountRegisters;
+				int buffer_size = count * 8;
+				buffer = Marshal.AllocHGlobal (buffer_size);
+				TargetError result = mono_debugger_server_get_callback_frame (
+					server_handle, stack_pointer.Address, exact_match, buffer);
+				if (result == TargetError.NoCallbackFrame)
+					return null;
+				check_error (result);
+				long[] retval = new long [count];
+				Marshal.Copy (buffer, retval, 0, count);
+
+				return new Registers (arch, retval);
+			} finally {
+				if (buffer != IntPtr.Zero)
+					Marshal.FreeHGlobal (buffer);
+			}
+		}
+
+		internal void SetRuntimeInfo (IntPtr mono_runtime_info)
+		{
+			mono_debugger_server_set_runtime_info (server_handle, mono_runtime_info);
 		}
 
 		internal struct ServerStackFrame
@@ -1244,7 +1387,7 @@ namespace Mono.Debugger.Backends
 		{
 			TargetAddress address, stack, frame;
 
-			internal StackFrame (TargetInfo info, ServerStackFrame frame)
+			internal StackFrame (TargetMemoryInfo info, ServerStackFrame frame)
 			{
 				this.address = new TargetAddress (info.AddressDomain, frame.Address);
 				this.stack = new TargetAddress (info.AddressDomain, frame.StackPointer);
@@ -1282,16 +1425,27 @@ namespace Mono.Debugger.Backends
 		{
 			check_disposed ();
 			ServerStackFrame frame;
-			TargetError result = mono_debugger_server_get_frame (
-				server_handle, out frame);
+			TargetError result = mono_debugger_server_get_frame (server_handle, out frame);
 			check_error (result);
 			return frame;
 		}
 
+		public StackFrame GetCurrentFrame (bool may_fail)
+		{
+			check_disposed ();
+			ServerStackFrame frame;
+			TargetError result = mono_debugger_server_get_frame (server_handle, out frame);
+			if (result == TargetError.None)
+				return new StackFrame (target_info, frame);
+			else if (may_fail)
+				return null;
+			else
+				throw new TargetException (result);
+		}
+
 		public StackFrame GetCurrentFrame ()
 		{
-			ServerStackFrame frame = get_current_frame ();
-			return new StackFrame (target_info, frame);
+			return GetCurrentFrame (false);
 		}
 
 		public TargetMemoryArea[] GetMemoryMaps ()
@@ -1364,6 +1518,10 @@ namespace Mono.Debugger.Backends
 			// child_event (ChildEventType.CHILD_MEMORY_CHANGED, 0);
 		}
 
+		public bool HasSignals {
+			get { return has_signals; }
+		}
+
 		public int SIGKILL {
 			get {
 				if (!has_signals || (signal_info.SIGKILL < 0))
@@ -1400,6 +1558,32 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
+		public bool IsManagedSignal (int signal)
+		{
+			if (!has_signals)
+				throw new InvalidOperationException ();
+
+			if ((signal == signal_info.SIGFPE) || (signal == signal_info.SIGQUIT) ||
+			    (signal == signal_info.SIGABRT) || (signal == signal_info.SIGSEGV) ||
+			    (signal == signal_info.SIGILL) || (signal == signal_info.SIGBUS))
+				return true;
+
+			return false;
+		}
+
+		/*
+		 * CAUTION: This is the hard limit of the Linux kernel, not the first
+		 *          user-visible real-time signal !
+		 */
+		public int Kernel_SIGRTMIN {
+			get {
+				if (!has_signals || (signal_info.Kernel_SIGRTMIN < 0))
+					throw new InvalidOperationException ();
+
+				return signal_info.Kernel_SIGRTMIN;
+			}
+		}
+
 		public int MonoThreadAbortSignal {
 			get {
 				if (!has_signals || (signal_info.MonoThreadAbortSignal < 0))
@@ -1427,13 +1611,6 @@ namespace Mono.Debugger.Backends
 			if (!this.disposed) {
 				// If this is a call to Dispose,
 				// dispose all managed resources.
-				if (disposing) {
-					if (bfd_disassembler != null) {
-						bfd_disassembler.Dispose ();
-						bfd_disassembler = null;
-					}
-				}
-				
 				this.disposed = true;
 
 				// Release unmanaged resources
