@@ -6,31 +6,32 @@ using System.Runtime.InteropServices;
 
 using Mono.Debugger;
 using Mono.Debugger.Languages;
+using Mono.Debugger.Architectures;
 
-namespace Mono.Debugger.Backends
+namespace Mono.Debugger.Backend
 {
 	internal delegate void BfdDisposedHandler (Bfd bfd);
  
-	internal class Bfd : SymbolFile, ISymbolContainer, ILanguageBackend
+	internal class Bfd : SymbolFile, ISymbolContainer
 	{
 		IntPtr bfd;
 		protected Module module;
 		protected BfdContainer container;
-		protected TargetInfo info;
+		protected TargetMemoryInfo info;
 		protected Bfd main_bfd;
-		protected Architecture arch;
 		TargetAddress first_link_map = TargetAddress.Null;
 		TargetAddress dynlink_breakpoint = TargetAddress.Null;
 		TargetAddress rdebug_state_addr = TargetAddress.Null;
 		TargetAddress entry_point = TargetAddress.Null;
 		bool is_loaded;
-		Hashtable load_handlers;
 		Hashtable symbols;
+		Hashtable local_symbols;
 		ArrayList simple_symbols;
 		BfdSymbolTable simple_symtab;
 		DwarfReader dwarf;
 		DwarfFrameReader frame_reader, eh_frame_reader;
 		bool dwarf_loaded;
+		bool frames_loaded;
 		bool has_debugging_info;
 		string filename, target;
 		bool is_coredump;
@@ -65,7 +66,7 @@ namespace Mono.Debugger.Backends
 
 				this.name = bfd_glue_get_section_name (section);
 				this.vma = bfd_glue_get_section_vma (section);
-				this.size = bfd_glue_get_section_size (section, true);
+				this.size = bfd_glue_get_section_size (section);
 				this.flags = bfd_glue_get_section_flags (section);
 
 				contents = new ObjectCache (
@@ -74,7 +75,7 @@ namespace Mono.Debugger.Backends
 
 			object get_section_contents (object user_data)
 			{
-				byte[] data = bfd.GetSectionContents (section, true);
+				byte[] data = bfd.GetSectionContents (section);
 				if (data == null)
 					throw new SymbolTableException ("Can't get bfd section {0}", name);
 				return new TargetReader (data, bfd.info);
@@ -83,7 +84,7 @@ namespace Mono.Debugger.Backends
 			public TargetReader GetReader (TargetAddress address)
 			{
 				TargetReader reader = (TargetReader) contents.Data;
-				reader.Offset = address.Address - vma;
+				reader.Offset = address.Address - bfd.BaseAddress.Address - vma;
 				return reader;
 			}
 
@@ -103,6 +104,9 @@ namespace Mono.Debugger.Backends
 		[DllImport("monodebuggerserver")]
 		extern static bool bfd_close (IntPtr bfd);
 
+		[DllImport("monodebuggerserver")]
+		extern static string bfd_glue_get_errormsg ();
+
 		[DllImport("monodebuggerserver")]
 		extern static string bfd_glue_get_target_name (IntPtr bfd);
 
@@ -146,7 +150,7 @@ namespace Mono.Debugger.Backends
 		extern static string bfd_glue_get_section_name (IntPtr section);
 
 		[DllImport("monodebuggerserver")]
-		extern static int bfd_glue_get_section_size (IntPtr section, bool raw_section);
+		extern static int bfd_glue_get_section_size (IntPtr section);
 
 		[DllImport("monodebuggerserver")]
 		extern static SectionFlags bfd_glue_get_section_flags (IntPtr section);
@@ -166,7 +170,7 @@ namespace Mono.Debugger.Backends
 			bfd_init ();
 		}
 
-		public Bfd (BfdContainer container, TargetInfo info, string filename,
+		public Bfd (BfdContainer container, TargetMemoryInfo info, string filename,
 			    Bfd main_bfd, TargetAddress base_address, bool is_loaded)
 		{
 			this.container = container;
@@ -176,8 +180,6 @@ namespace Mono.Debugger.Backends
 			this.main_bfd = main_bfd;
 			this.is_loaded = is_loaded;
 
-			load_handlers = new Hashtable ();
-
 			bfd = bfd_glue_openr (filename, null);
 			if (bfd == IntPtr.Zero)
 				throw new SymbolTableException ("Can't read symbol file: {0}", filename);
@@ -191,13 +193,9 @@ namespace Mono.Debugger.Backends
 
 			target = bfd_glue_get_target_name (bfd);
 			if ((target == "elf32-i386") || (target == "elf64-x86-64")) {
-				if (target == "elf32-i386")
-					arch = new Architecture_I386 (container.Process, info);
-				else
-					arch = new Architecture_X86_64 (container.Process, info);
-
 				if (!is_coredump) {
 					Section text = GetSectionByName (".text", true);
+					Section bss = GetSectionByName (".bss", true);
 
 					if (!base_address.IsNull)
 						start_address = new TargetAddress (
@@ -206,10 +204,14 @@ namespace Mono.Debugger.Backends
 					else
 						start_address = new TargetAddress (
 							info.AddressDomain, text.vma);
-					end_address = start_address + text.size;
-				} else {
-					start_address = main_bfd.StartAddress;
-					end_address = main_bfd.EndAddress;
+
+					if (!base_address.IsNull)
+						end_address = new TargetAddress (
+							info.AddressDomain,
+							base_address.Address + bss.vma + bss.size);
+					else
+						end_address = new TargetAddress (
+							info.AddressDomain, bss.vma + bss.size);
 				}
 
 				read_bfd_symbols ();
@@ -261,6 +263,7 @@ namespace Mono.Debugger.Backends
 			int num_symbols = bfd_glue_get_symbols (bfd, out symtab);
 
 			symbols = new Hashtable ();
+			local_symbols = new Hashtable ();
 			simple_symbols = new ArrayList ();
 
 			for (int i = 0; i < num_symbols; i++) {
@@ -278,9 +281,8 @@ namespace Mono.Debugger.Backends
 					symbols.Add (name, relocated);
 				else if ((main_bfd == null) && name.StartsWith ("MONO_DEBUGGER__"))
 					symbols.Add (name, relocated);
-				else if (name.StartsWith ("__pthread_") ||
-					 (name == "__libc_pthread_functions"))
-					symbols.Add (name, relocated);
+				else if (!local_symbols.Contains (name))
+					local_symbols.Add (name, relocated);
 
 				simple_symbols.Add (new Symbol (name, relocated, 0));
 			}
@@ -323,6 +325,20 @@ namespace Mono.Debugger.Backends
 			return false;
 		}
 
+		void insert_dynlink_breakpoint (Inferior inferior)
+		{
+			Instruction insn = Architecture.ReadInstruction (
+				inferior, dynlink_breakpoint);
+			if ((insn == null) || !insn.CanInterpretInstruction) {
+				Console.WriteLine ("UNKNOWN DYNLINK BREAKPOINT: {0}",
+						   dynlink_breakpoint);
+				return;
+			}
+
+			AddressBreakpoint dynlink_bpt = new DynlinkBreakpoint (this, insn);
+			dynlink_bpt.Insert (inferior);
+		}
+
 		bool read_dynamic_info (Inferior inferior, TargetMemoryAccess target)
 		{
 			if (initialized)
@@ -370,9 +386,7 @@ namespace Mono.Debugger.Backends
 				return false;
 
 			if (inferior != null)
-				inferior.BreakpointManager.InsertBreakpoint (
-					inferior, new DynlinkBreakpoint (this, dynlink_breakpoint),
-					dynlink_breakpoint);
+				insert_dynlink_breakpoint (inferior);
 
 			has_shlib_info = true;
 			return true;
@@ -452,7 +466,7 @@ namespace Mono.Debugger.Backends
 
 		public Architecture Architecture {
 			get {
-				return arch;
+				return container.Architecture;
 			}
 		}
 
@@ -472,7 +486,7 @@ namespace Mono.Debugger.Backends
 					info.AddressDomain, BaseAddress.Address + address);
 		}
 
-		public TargetInfo TargetInfo {
+		public TargetMemoryInfo TargetMemoryInfo {
 			get {
 				return info;
 			}
@@ -529,12 +543,6 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		internal override ILanguageBackend LanguageBackend {
-			get {
-				return this;
-			}
-		}
-
 		internal DwarfReader DwarfReader {
 			get { return dwarf; }
 		}
@@ -552,7 +560,7 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		public override SourceMethod[] GetMethods (SourceFile file)
+		public override MethodSource[] GetMethods (SourceFile file)
 		{
 			if (dwarf != null)
 				return dwarf.GetMethods (file);
@@ -560,20 +568,7 @@ namespace Mono.Debugger.Backends
 			throw new InvalidOperationException ();
 		}
 
-		public override Method GetMethod (int domain, long handle)
-		{
-			if (dwarf != null)
-				return dwarf.GetMethod (handle);
-
-			throw new InvalidOperationException ();
-		}
-
-		public override TargetFunctionType LookupMethod (string class_name, string method_name)
-		{
-			return null;
-		}
-
-		public override SourceMethod FindMethod (string name)
+		public override MethodSource FindMethod (string name)
 		{
 			if (dwarf != null)
 				return dwarf.FindMethod (name);
@@ -607,7 +602,7 @@ namespace Mono.Debugger.Backends
 			long vma_base = base_address.IsNull ? 0 : base_address.Address;
 			Section section = GetSectionByName (".debug_frame", false);
 			if (section != null) {
-				byte[] contents = GetSectionContents (section.section, false);
+				byte[] contents = GetSectionContents (section.section);
 				TargetBlob blob = new TargetBlob (contents, info);
 				frame_reader = new DwarfFrameReader (
 					this, blob, vma_base + section.vma, false);
@@ -615,16 +610,16 @@ namespace Mono.Debugger.Backends
 
 			section = GetSectionByName (".eh_frame", false);
 			if (section != null) {
-				byte[] contents = GetSectionContents (section.section, false);
+				byte[] contents = GetSectionContents (section.section);
 				TargetBlob blob = new TargetBlob (contents, info);
 				eh_frame_reader = new DwarfFrameReader (
 					this, blob, vma_base + section.vma, true);
 			}
 		}
 
-		void load_dwarf ()
+		void load_frames ()
 		{
-			if (dwarf_loaded)
+			if (frames_loaded)
 				return;
 
 			try {
@@ -634,12 +629,16 @@ namespace Mono.Debugger.Backends
 						   "symbol file `{0}': {1}", FileName, ex);
 			}
 
-			if (!has_debugging_info)
+			frames_loaded = true;
+		}
+
+		void load_dwarf ()
+		{
+			if (dwarf_loaded || !has_debugging_info)
 				return;
 
 			try {
-				dwarf = new DwarfReader (
-					this, module, container.Process.SourceFileFactory);
+				dwarf = new DwarfReader (this, module);
 			} catch (Exception ex) {
 				Console.WriteLine ("Cannot read DWARF debugging info from " +
 						   "symbol file `{0}': {1}", FileName, ex);
@@ -649,9 +648,8 @@ namespace Mono.Debugger.Backends
 
 			dwarf_loaded = true;
 
-			if (dwarf != null) {
+			if (dwarf != null)
 				has_debugging_info = true;
-			}
 		}
 
 		void unload_dwarf ()
@@ -659,15 +657,14 @@ namespace Mono.Debugger.Backends
 			if (!dwarf_loaded || !has_debugging_info)
 				return;
 
-			frame_reader = null;
-			eh_frame_reader = null;
-
 			dwarf_loaded = false;
 			dwarf = null;
 		}
 
 		internal override void OnModuleChanged ()
 		{
+			load_frames ();
+
 			if (module.LoadSymbols) {
 				load_dwarf ();
 			} else {
@@ -675,16 +672,6 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		internal BfdDisassembler GetDisassembler (TargetMemoryAccess memory)
-		{
-			IntPtr dis = disassembler (bfd);
-
-			IntPtr info = bfd_glue_init_disassembler (bfd);
-
-			return new BfdDisassembler (
-				container.Process.SymbolTableManager, memory, dis, info);
-		}
-
 		public TargetAddress this [string name] {
 			get {
 				if (symbols == null)
@@ -697,22 +684,41 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		internal Section this [long address] {
-			get {
-				read_sections ();
-				foreach (Section section in sections) {
-					if ((address < section.vma) || (address >= section.vma + section.size))
-						continue;
+		public TargetAddress LookupLocalSymbol (string name)
+		{
+			if (local_symbols == null)
+				return TargetAddress.Null;
+
+			if (local_symbols.Contains (name))
+				return (TargetAddress) local_symbols [name];
+
+			return TargetAddress.Null;
+		}
+
+		internal Section FindSection (long address)
+		{
+			read_sections ();
+			foreach (Section section in sections) {
+				long relocated = base_address.Address + section.vma;
+
+				if ((address < relocated) || (address >= relocated + section.size))
+					continue;
 
 					return section;
-				}
+			}
 
-				if (main_bfd != null)
-					return main_bfd [address];
+			return null;
+		}
 
-				throw new SymbolTableException (String.Format (
+		internal Section this [long address] {
+			get {
+				Section section = FindSection (address);
+				if (section != null)
+					return section;
+
+				throw new SymbolTableException (
 					"No section in file {1} contains address {0:x}.", address,
-					filename));
+					filename);
 			}
 		}
 
@@ -771,10 +777,18 @@ namespace Mono.Debugger.Backends
 			return maps;
 		}
 
-		public TargetReader GetReader (TargetAddress address)
+		public TargetReader GetReader (TargetAddress address, bool may_fail)
 		{
-			Section section = this [address.Address];
-			return section.GetReader (address);
+			Section section = FindSection (address.Address);
+			if (section != null)
+				return section.GetReader (address);
+
+			if (may_fail)
+				return null;
+
+			throw new SymbolTableException (
+				"No section in file {1} contains address {0:x}.", address,
+				filename);
 		}
 
 		public TargetAddress GetSectionAddress (string name)
@@ -789,36 +803,41 @@ namespace Mono.Debugger.Backends
 			return GetAddress (vma);
 		}
 
-		public byte[] GetSectionContents (string name, bool raw_section)
+		public byte[] GetSectionContents (string name)
 		{
 			IntPtr section;
 
 			section = bfd_get_section_by_name (bfd, name);
 			if (section == IntPtr.Zero)
-				throw new SymbolTableException ("Can't get bfd section {0}", name);
+				throw new SymbolTableException ("Can't find bfd section {0}", name);
 
-			byte[] contents = GetSectionContents (section, raw_section);
+			byte[] contents = GetSectionContents (section);
 			if (contents == null)
-				throw new SymbolTableException ("Can't get bfd section {0}", name);
+				throw new SymbolTableException ("Can't read bfd section {0}", name);
 			return contents;
 		}
 
-		public TargetReader GetSectionReader (string name, bool raw_section)
+		public TargetReader GetSectionReader (string name)
 		{
-			byte[] contents = GetSectionContents (name, raw_section);
+			byte[] contents = GetSectionContents (name);
 			if (contents == null)
-				throw new SymbolTableException ("Can't get bfd section {0}", name);
+				throw new SymbolTableException ("Can't read bfd section {0}", name);
 			return new TargetReader (contents, info);
 		}
 
-		byte[] GetSectionContents (IntPtr section, bool raw_section)
+		byte[] GetSectionContents (IntPtr section)
 		{
-			int size = bfd_glue_get_section_size (section, raw_section);
+			int size = bfd_glue_get_section_size (section);
 			IntPtr data = IntPtr.Zero;
 			try {
 				data = Marshal.AllocHGlobal (size);
-				if (!bfd_glue_get_section_contents (bfd, section, data, size))
-					return null;
+				if (!bfd_glue_get_section_contents (bfd, section, data, size)) {
+					string error = bfd_glue_get_errormsg ();
+					string name = bfd_glue_get_section_name (section);
+
+					throw new SymbolTableException (
+						"Can't read bfd section {0}: {1}", name, error);
+				}
 				byte[] retval = new byte [size];
 				Marshal.Copy (data, retval, 0, size);
 				return retval;
@@ -893,7 +912,7 @@ namespace Mono.Debugger.Backends
 
 		public bool IsContinuous {
 			get {
-				return !end_address.IsNull;
+				return !is_coredump && !end_address.IsNull;
 			}
 		}
 
@@ -915,81 +934,41 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		//
-		// ILanguageBackend
-		//
-
-		string ILanguageBackend.Name {
-			get {
-				return "Native";
-			}
-		}
-
-		TargetAddress ILanguageBackend.RuntimeInvokeFunc {
-			get {
-				throw new InvalidOperationException ();
-			}
-		}
-
-		TargetAddress ILanguageBackend.CompileMethodFunc {
-			get {
-				return TargetAddress.Null;
-			}
-		}
-
-		TargetAddress ILanguageBackend.GetVirtualMethodFunc {
-			get {
-				throw new InvalidOperationException ();
-			}
-		}
-
-		TargetAddress ILanguageBackend.GetBoxedObjectFunc {
-			get {
-				throw new InvalidOperationException ();
-			}
-		}
-
-		TargetAddress ILanguageBackend.GetTrampolineAddress (TargetMemoryAccess memory,
-								     TargetAddress address,
-								     out bool is_start)
-		{
-			return GetTrampoline (memory, address, out is_start);
-		}
-
-		SourceMethod ILanguageBackend.GetTrampoline (TargetMemoryAccess memory,
-							     TargetAddress address)
-		{
-			return null;
-		}
-
-		void ILanguageBackend.Notification (Inferior inferior, NotificationType type,
-						    TargetAddress data, long arg)
-		{ }
-
-		public TargetAddress GetTrampoline (TargetMemoryAccess memory,
-						    TargetAddress address, out bool is_start)
+		internal bool GetTrampoline (TargetMemoryAccess memory, TargetAddress address,
+					     out TargetAddress trampoline, out bool is_start)
 		{
 			if (!has_got || (address < plt_start) || (address > plt_end)) {
 				is_start = false;
-				return TargetAddress.Null;
+				trampoline = TargetAddress.Null;
+				return false;
 			}
 
-			int insn_size;
-			TargetAddress jmp_target = arch.GetJumpTarget (memory, address, out insn_size);
-			if (jmp_target.IsNull) {
+			Instruction target_insn = container.Architecture.ReadInstruction (
+				memory, address);
+			if ((target_insn == null) || !target_insn.HasInstructionSize ||
+			    ((target_insn.InstructionType != Instruction.Type.Jump) &&
+			     (target_insn.InstructionType != Instruction.Type.IndirectJump))) {
 				is_start = false;
-				return TargetAddress.Null;
+				trampoline = TargetAddress.Null;
+				return false;
 			}
 
-			TargetAddress method = memory.ReadAddress (jmp_target);
-
-			if (method != address + 6) {
+			TargetAddress call_target = target_insn.GetEffectiveAddress (memory);
+			if (call_target.IsNull) {
 				is_start = false;
-				return method;
+				trampoline = TargetAddress.Null;
+				return false;
+			}
+
+			if (call_target != address + target_insn.InstructionSize) {
+				is_start = false;
+				trampoline = call_target;
+				return true;
 			}
 
 			is_start = true;
-			return memory.ReadAddress (got_start + 2 * info.TargetAddressSize);
+			trampoline = memory.ReadAddress (got_start + 3 * info.TargetAddressSize);
+			return true;
 		}
 
 		internal override StackFrame UnwindStack (StackFrame frame, TargetMemoryAccess memory)
@@ -999,7 +978,7 @@ namespace Mono.Debugger.Backends
 
 			StackFrame new_frame;
 			try {
-				new_frame = arch.TrySpecialUnwind (frame, memory);
+				new_frame = container.Architecture.TrySpecialUnwind (frame, memory);
 				if (new_frame != null)
 					return new_frame;
 			} catch {
@@ -1007,13 +986,15 @@ namespace Mono.Debugger.Backends
 
 			try {
 				if (frame_reader != null) {
-					new_frame = frame_reader.UnwindStack (frame, memory, arch);
+					new_frame = frame_reader.UnwindStack (
+						frame, memory, container.Architecture);
 					if (new_frame != null)
 						return new_frame;
 				}
 
 				if (eh_frame_reader != null) {
-					new_frame = eh_frame_reader.UnwindStack (frame, memory, arch);
+					new_frame = eh_frame_reader.UnwindStack (
+						frame, memory, container.Architecture);
 					if (new_frame != null)
 						return new_frame;
 				}
@@ -1030,23 +1011,6 @@ namespace Mono.Debugger.Backends
 				dwarf.ReadTypes ();
 		}
 
-		internal override ILoadHandler RegisterLoadHandler (Thread target,
-								    SourceMethod method,
-								    MethodLoadedHandler handler,
-								    object user_data)
-		{
-			LoadHandlerData data = new LoadHandlerData (
-				this, method, handler, user_data);
-
-			load_handlers.Add (data, true);
-			return data;
-		}
-
-		protected void UnRegisterLoadHandler (LoadHandlerData data)
-		{
-			load_handlers.Remove (data);
-		}
-
 		void module_loaded (Inferior inferior, TargetAddress address)
 		{
 			this.base_address = address;
@@ -1066,46 +1030,17 @@ namespace Mono.Debugger.Backends
 				TargetAddress nptl_setxid = inferior.ReadAddress (
 					vtable + 51 * info.TargetAddressSize);
 
-				if (!nptl_setxid.IsNull)
-					inferior.BreakpointManager.InsertBreakpoint (
-						inferior, new SetXidBreakpoint (this, nptl_setxid),
-						nptl_setxid);
+				if (!nptl_setxid.IsNull) {
+					AddressBreakpoint setxid_bpt = new SetXidBreakpoint (
+						this, nptl_setxid);
+					setxid_bpt.Insert (inferior);
+				}
 			}
 
 			if (dwarf != null) {
 				dwarf.ModuleLoaded ();
 				has_debugging_info = true;
 			}
-
-			foreach (LoadHandlerData data in load_handlers.Keys)
-				data.Handler (inferior, data.Method, data.UserData);
-		}
-
-		protected sealed class LoadHandlerData : ILoadHandler
-		{
-			public readonly Bfd Bfd;
-			public readonly SourceMethod Method;
-			public readonly MethodLoadedHandler Handler;
-			public readonly object UserData;
-
-			public LoadHandlerData (Bfd bfd, SourceMethod method,
-						MethodLoadedHandler handler,
-						object user_data)
-			{
-				this.Bfd = bfd;
-				this.Method = method;
-				this.Handler = handler;
-				this.UserData = user_data;
-			}
-
-			object ILoadHandler.UserData {
-				get { return UserData; }
-			}
-
-			public void Remove ()
-			{
-				Bfd.UnRegisterLoadHandler (this);
-			}
 		}
 
 		//
@@ -1116,18 +1051,16 @@ namespace Mono.Debugger.Backends
 		{
 			Bfd bfd;
 			Symbol[] list;
-			TargetAddress start, end;
 
 			public BfdSymbolTable (Bfd bfd)
 			{
 				this.bfd = bfd;
-				this.start = bfd.StartAddress;
-				this.end = bfd.EndAddress;
 			}
 
 			public Symbol SimpleLookup (TargetAddress address, bool exact_match)
 			{
-				if ((address < start) || (address >= end))
+				if (bfd.IsContinuous &&
+				    ((address < bfd.StartAddress) || (address >= bfd.EndAddress)))
 					return null;
 
 				if (list == null) {
@@ -1186,7 +1119,7 @@ namespace Mono.Debugger.Backends
 			internal override bool BreakpointHandler (Inferior inferior,
 								  out bool remain_stopped)
 			{
-				bfd.arch.Hack_ReturnNull (inferior);
+				bfd.Architecture.Hack_ReturnNull (inferior);
 				remain_stopped = false;
 				return true;
 			}
@@ -1195,11 +1128,13 @@ namespace Mono.Debugger.Backends
 		protected class DynlinkBreakpoint : AddressBreakpoint
 		{
 			protected readonly Bfd bfd;
+			public readonly Instruction instruction;
 
-			public DynlinkBreakpoint (Bfd bfd, TargetAddress address)
-				: base ("dynlink", ThreadGroup.System, address)
+			public DynlinkBreakpoint (Bfd bfd, Instruction instruction)
+				: base ("dynlink", ThreadGroup.System, instruction.Address)
 			{
 				this.bfd = bfd;
+				this.instruction = instruction;
 			}
 
 			public override bool CheckBreakpointHit (Thread target, TargetAddress address)
@@ -1211,6 +1146,8 @@ namespace Mono.Debugger.Backends
 								  out bool remain_stopped)
 			{
 				bfd.dynlink_handler (inferior);
+				if (!instruction.InterpretInstruction (inferior))
+					throw new InternalError ();
 				remain_stopped = false;
 				return true;
 			}
