@@ -28,7 +28,8 @@ typedef enum {
 	COMMAND_ERROR_DR_OCCUPIED,
 	COMMAND_ERROR_MEMORY_ACCESS,
 	COMMAND_ERROR_NOT_IMPLEMENTED,
-	COMMAND_ERROR_IO_ERROR
+	COMMAND_ERROR_IO_ERROR,
+	COMMAND_ERROR_NO_CALLBACK_FRAME,
 } ServerCommandError;
 
 typedef enum {
@@ -60,21 +61,48 @@ typedef struct {
 	guint64 frame_address;
 } StackFrame;
 
+#define EXECUTABLE_CODE_CHUNK_SIZE		16
+
+typedef struct
+{
+	guint32 address_size;
+	guint64 notification_address;
+	guint64 executable_code_buffer;
+	guint32 executable_code_buffer_size;
+	guint32 executable_code_chunk_size;
+	guint32 executable_code_total_chunks;
+	guint64 breakpoint_info_area;
+	guint64 breakpoint_table;
+	guint32 breakpoint_table_size;
+
+	/* Private */
+	guint8 *breakpoint_table_bitfield;
+	guint8 *executable_code_bitfield;
+} MonoRuntimeInfo;
+
 /* This is an opaque data structure which the backend may use to store stuff. */
 typedef struct InferiorVTable InferiorVTable;
 typedef struct InferiorHandle InferiorHandle;
 typedef struct ServerHandle ServerHandle;
 typedef struct ArchInfo ArchInfo;
 
+typedef struct IOThreadData IOThreadData;
+
 /* C# delegates. */
-typedef void (*ChildOutputFunc) (const char *output);
+typedef void (*ChildOutputFunc) (gboolean is_stderr, const char *output);
 
 typedef struct {
 	int sigkill;
 	int sigstop;
 	int sigint;
 	int sigchld;
-
+	int sigfpe;
+	int sigquit;
+	int sigabrt;
+	int sigsegv;
+	int sigill;
+	int sigbus;
+	int kernel_sigrtmin;
 	int mono_thread_abort;
 } SignalInfo;
 
@@ -92,6 +120,7 @@ typedef struct {
 struct ServerHandle {
 	ArchInfo *arch;
 	InferiorHandle *inferior;
+	MonoRuntimeInfo *mono_runtime;
 	BreakpointManager *bpm;
 };
 
@@ -105,13 +134,18 @@ struct InferiorVTable {
 	ServerCommandError    (* initialize_thread)   (ServerHandle       *handle,
 						       guint32             pid);
 
+	void                  (* set_runtime_info)    (ServerHandle       *handle,
+						       MonoRuntimeInfo    *mono_runtime_info);
+
+	void                  (* io_thread_main)      (IOThreadData       *io_data,
+						       ChildOutputFunc     func);
+
 	ServerCommandError    (* spawn)               (ServerHandle       *handle,
 						       const gchar        *working_directory,
 						       const gchar       **argv,
 						       const gchar       **envp,
 						       gint               *child_pid,
-						       ChildOutputFunc     stdout_handler,
-						       ChildOutputFunc     stderr_handler,
+						       IOThreadData      **io_data,
 						       gchar             **error);
 
 	ServerCommandError    (* attach)              (ServerHandle       *handle,
@@ -130,7 +164,12 @@ struct InferiorVTable {
 						       guint32              status,
 						       guint64             *arg,
 						       guint64             *data1,
-						       guint64             *data2);
+						       guint64             *data2,
+						       guint32             *opt_data_size,
+						       gpointer            *opt_data);
+
+	ServerStatusMessageType (* dispatch_simple)   (guint32              status,
+						       guint32             *arg);
 
 	/* Get sizeof (int), sizeof (long) and sizeof (void *) from the target. */
 	ServerCommandError    (* get_target_info)     (guint32            *target_int_size,
@@ -152,6 +191,8 @@ struct InferiorVTable {
 	 */
 	ServerCommandError    (* step)                (ServerHandle     *handle);
 
+	ServerCommandError    (* resume)              (ServerHandle     *handle);
+
 	/*
 	 * Get the current program counter.
 	 * Return COMMAND_ERROR_NOT_STOPPED if the target is currently running.
@@ -209,12 +250,23 @@ struct InferiorVTable {
 	ServerCommandError    (* call_method_1)       (ServerHandle     *handle,
 						       guint64           method,
 						       guint64           method_argument,
+						       guint64           data_argument,
+						       guint64           data_argument2,
 						       const gchar      *string_argument,
 						       guint64           callback_argument);
 
 	ServerCommandError    (* call_method_2)       (ServerHandle     *handle,
 						       guint64           method,
+						       guint32           data_size,
+						       gconstpointer     data_buffer,
+						       guint64           callback_argument);
+
+	ServerCommandError    (* call_method_3)       (ServerHandle     *handle,
+						       guint64           method_address,
 						       guint64           method_argument,
+						       guint64           address_argument,
+						       guint32           blob_size,
+						       gconstpointer     blob_data,
 						       guint64           callback_argument);
 
 	ServerCommandError    (* call_method_invoke)  (ServerHandle     *handle,
@@ -223,12 +275,20 @@ struct InferiorVTable {
 						       guint32           num_params,
 						       guint32           glob_size,
 						       guint64          *param_data,
-						       guint32          *offset_data,
+						       gint32           *offset_data,
 						       gconstpointer     blob_data,
 						       guint64           callback_argument,
 						       gboolean          debug);
 
-	ServerCommandError    (* abort_invoke)        (ServerHandle     *handle);
+	ServerCommandError    (* execute_instruction) (ServerHandle     *handle,
+						       const guint8     *instruction,
+						       guint32           size,
+						       gboolean          update_ip);
+
+	ServerCommandError    (* mark_rti_frame)      (ServerHandle     *handle);
+
+	ServerCommandError    (* abort_invoke)        (ServerHandle     *handle,
+						       guint64           stack_pointer);
 
 	/*
 	 * Insert a breakpoint at address `address' in the target's address space.
@@ -311,9 +371,6 @@ struct InferiorVTable {
 	ServerCommandError    (* get_signal_info)     (ServerHandle     *handle,
 						       SignalInfo      **sinfo);
 
-	void                  (* set_notification)    (ServerHandle     *handle,
-						       guint64           notification);
-
 	ServerCommandError    (* get_threads)         (ServerHandle     *handle,
 						       guint32          *count,
 						       guint32         **threads);
@@ -322,14 +379,22 @@ struct InferiorVTable {
 						       gchar           **exe_file,
 						       gchar           **cwd,
 						       guint32          *nargs,
-						       gchar           **cmdline_args);
+						       gchar          ***cmdline_args);
 
-	ServerCommandError    (* init_after_fork)     (ServerHandle      *handle);
+	ServerCommandError    (* detach_after_fork)   (ServerHandle      *handle);
 
 	ServerCommandError    (* push_registers)      (ServerHandle      *handle,
 						       guint64           *new_rsp);
 
 	ServerCommandError    (* pop_registers)       (ServerHandle      *handle);
+
+	ServerCommandError    (* get_callback_frame)  (ServerHandle      *handle,
+						       guint64            stack_pointer,
+						       gboolean           exact_match,
+						       guint64           *registers);
+
+	void         (* get_registers_from_core_file) (guint64           *values,
+						       const guint8      *buffer);
 };
 
 /*
@@ -355,14 +420,17 @@ ServerCommandError
 mono_debugger_server_initialize_thread    (ServerHandle       *handle,
 					   guint32             pid);
 
+void
+mono_debugger_server_io_thread_main       (IOThreadData       *io_data,
+					   ChildOutputFunc     func);
+
 ServerCommandError
 mono_debugger_server_spawn                (ServerHandle       *handle,
 					   const gchar        *working_directory,
 					   const gchar       **argv,
 					   const gchar       **envp,
 					   gint               *child_pid,
-					   ChildOutputFunc     stdout_handler,
-					   ChildOutputFunc     stderr_handler,
+					   IOThreadData      **io_data,
 					   gchar             **error);
 
 ServerCommandError
@@ -380,7 +448,9 @@ mono_debugger_server_dispatch_event       (ServerHandle            *handle,
 					   guint32                  status,
 					   guint64                 *arg,
 					   guint64                 *data1,
-					   guint64                 *data2);
+					   guint64                 *data2,
+					   guint32                 *opt_data_size,
+					   gpointer                *opt_data);
 
 ServerCommandError
 mono_debugger_server_get_target_info      (guint32            *target_int_size,
@@ -402,6 +472,9 @@ mono_debugger_server_step                 (ServerHandle       *handle);
 ServerCommandError
 mono_debugger_server_continue             (ServerHandle       *handle);
 
+ServerCommandError
+mono_debugger_server_resume               (ServerHandle       *handle);
+
 ServerCommandError
 mono_debugger_server_detach               (ServerHandle       *handle);
 
@@ -433,13 +506,25 @@ ServerCommandError
 mono_debugger_server_call_method_1        (ServerHandle       *handle,
 					   guint64             method_address,
 					   guint64             method_argument,
+					   guint64             data_argument,
+					   guint64             data_argument2,
 					   const gchar        *string_argument,
 					   guint64             callback_argument);
 
 ServerCommandError
 mono_debugger_server_call_method_2        (ServerHandle       *handle,
+					   guint64             method_address,
+					   guint32             data_size,
+					   gconstpointer       data_buffer,
+					   guint64             callback_argument);
+
+ServerCommandError
+mono_debugger_server_call_method_3        (ServerHandle       *handle,
 					   guint64             method_address,
 					   guint64             method_argument,
+					   guint64             address_argument,
+					   guint32             blob_size,
+					   gconstpointer       blob_data,
 					   guint64             callback_argument);
 
 ServerCommandError
@@ -449,13 +534,23 @@ mono_debugger_server_call_method_invoke   (ServerHandle       *handle,
 					   guint32             num_params,
 					   guint32             blob_size,
 					   guint64            *param_data,
-					   guint32            *offset_data,
+					   gint32             *offset_data,
 					   gconstpointer       blob_data,
 					   guint64             callback_argument,
 					   gboolean            debug);
 
 ServerCommandError
-mono_debugger_server_abort_invoke        (ServerHandle        *handle);
+mono_debugger_execute_instruction         (ServerHandle        *handle,
+					   const guint8        *instruction,
+					   guint32              instruction_size,
+					   gboolean             update_ip);
+
+ServerCommandError
+mono_debugger_mark_rti_framenvoke        (ServerHandle        *handle);
+
+ServerCommandError
+mono_debugger_server_abort_invoke        (ServerHandle        *handle,
+					  guint64              stack_pointer);
 
 ServerCommandError
 mono_debugger_server_insert_breakpoint   (ServerHandle        *handle,
@@ -509,8 +604,8 @@ mono_debugger_server_get_signal_info     (ServerHandle        *handle,
 					  SignalInfo         **sinfo);
 
 void
-mono_debugger_server_set_notification    (ServerHandle        *handle,
-					  guint64              notification);
+mono_debugger_server_set_runtime_info    (ServerHandle        *handle,
+					  MonoRuntimeInfo     *mono_runtime);
 
 ServerCommandError
 mono_debugger_server_get_threads         (ServerHandle        *handle,
@@ -522,10 +617,10 @@ mono_debugger_server_get_application     (ServerHandle        *handle,
 					  gchar              **exe_file,
 					  gchar              **cwd,
 					  guint32             *nargs,
-					  gchar              **cmdline_args);
+					  gchar             ***cmdline_args);
 
 ServerCommandError
-mono_debugger_server_init_after_fork     (ServerHandle        *handle);
+mono_debugger_server_detach_after_fork   (ServerHandle        *handle);
 
 ServerCommandError
 mono_debugger_server_push_registers      (ServerHandle        *handle,
@@ -534,6 +629,33 @@ mono_debugger_server_push_registers      (ServerHandle        *handle,
 ServerCommandError
 mono_debugger_server_pop_registers       (ServerHandle        *handle);
 
+ServerCommandError
+mono_debugger_server_get_callback_frame  (ServerHandle        *handle,
+					  guint64              stack_pointer,
+					  gboolean             exact_match,
+					  guint64             *registers);
+
+MonoRuntimeInfo *
+mono_debugger_server_initialize_mono_runtime (guint32 address_size,
+					      guint64 notification_address,
+					      guint64 executable_code_buffer,
+					      guint32 executable_code_buffer_size,
+					      guint64 breakpoint_info_area,
+					      guint64 breakpoint_table,
+					      guint32 breakpoint_table_size);
+
+void
+mono_debugger_server_finalize_mono_runtime (MonoRuntimeInfo *runtime);
+
+void
+mono_debugger_server_initialize_code_buffer (MonoRuntimeInfo *runtime,
+					     guint64 executable_code_buffer,
+					     guint32 executable_code_buffer_size);
+
+void
+mono_debugger_server_get_registers_from_core_file (guint64 *values,
+						   const guint8 *buffer);
+
 /* POSIX semaphores */
 
 void mono_debugger_server_sem_init (void);
