@@ -1,11 +1,12 @@
 using System;
-using Math = System.Math;
 using System.Text;
 using System.IO;
 using System.Threading;
 using System.Reflection;
 using System.Collections;
+using System.Collections.Generic;
 using System.Globalization;
+using System.Runtime.Remoting;
 using System.Runtime.Serialization;
 using System.Runtime.InteropServices;
 using System.Runtime.Serialization.Formatters.Binary;
@@ -23,6 +24,8 @@ namespace Mono.Debugger.Frontend
 		DebuggerSession session;
 		DebuggerEngine engine;
 
+		SourceFileFactory source_factory;
+
 		Debugger debugger;
 		Process main_process;
 		Process current_process;
@@ -31,11 +34,13 @@ namespace Mono.Debugger.Frontend
 		StyleBase current_style;
 
 		bool is_interactive;
+		bool is_script;
 		int exit_code = 0;
 		int interrupt_level;
 
 		ExpressionParser parser;
 		ManualResetEvent interrupt_event;
+		ManualResetEvent process_event;
 		Thread current_thread;
 
 		internal static readonly string DirectorySeparatorStr;
@@ -51,11 +56,15 @@ namespace Mono.Debugger.Frontend
 		{
 			this.config = config;
 			this.is_interactive = is_interactive;
+			this.is_script = options.IsScript;
 			this.parser = new ExpressionParser (this);
 			this.session = new DebuggerSession (config, options, "main", parser);
 			this.engine = new DebuggerEngine (this);
 
+			source_factory = new SourceFileFactory ();
+
 			interrupt_event = new ManualResetEvent (false);
+			process_event = new ManualResetEvent (false);
 
 			styles = new Hashtable ();
 			styles.Add ("cli", new StyleCLI (this));
@@ -63,9 +72,6 @@ namespace Mono.Debugger.Frontend
 			current_style = (StyleBase) styles ["cli"];
 		}
 
-		AppDomain debugger_domain;
-		static int domain_age = 0;
-
 		public void Exit ()
 		{
 			try {
@@ -93,6 +99,10 @@ namespace Mono.Debugger.Frontend
 			get { return engine; }
 		}
 
+		public SourceFileFactory SourceFileFactory {
+			get { return source_factory; }
+		}
+
 		public Process CurrentProcess {
 			get {
 				if (current_process == null)
@@ -141,10 +151,12 @@ namespace Mono.Debugger.Frontend
 
 		public bool IsInteractive {
 			get { return is_interactive; }
+			set { is_interactive = value; }
 		}
 
 		public bool IsScript {
-			get { return Options.IsScript; }
+			get { return is_script; }
+			set { is_script = value; }
 		}
 
 		public DebuggerOptions Options {
@@ -220,9 +232,6 @@ namespace Mono.Debugger.Frontend
 			ArrayList list = new ArrayList ();
 
 			foreach (TargetFunctionType method in methods) {
-				if (method.Source == null)
-					continue;
-
 				list.Add (method);
 				Report.Print ("{0,4}  {1}\n", list.Count, method.Name);
 			}
@@ -257,6 +266,9 @@ namespace Mono.Debugger.Frontend
 
 		protected void CheckLastEvent (Thread thread)
 		{
+			if (thread == null)
+				return;
+
 			TargetEventArgs args = thread.GetLastTargetEvent ();
 			if (args == null)
 				return;
@@ -283,11 +295,7 @@ namespace Mono.Debugger.Frontend
 				current_process = main_process = debugger.Run (session);
 
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
-
-				// FIXME: Read the FIXME in DebuggerSession.cs !
-				session.MainProcessReachedMain (current_process);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException) {
@@ -313,10 +321,7 @@ namespace Mono.Debugger.Frontend
 
 				current_process = main_process = debugger.Attach (session, pid);
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
-
-				session.MainProcessReachedMain (current_process);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException) {
@@ -345,8 +350,6 @@ namespace Mono.Debugger.Frontend
 
 				current_thread = current_process.MainThread;
 
-				CheckLastEvent (current_thread);
-
 				return current_process;
 			} catch (TargetException) {
 				debugger.Dispose ();
@@ -375,19 +378,16 @@ namespace Mono.Debugger.Frontend
 				current_process = main_process = debugger.Run (session);
 
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
-
-				session.MainProcessReachedMain (current_process);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException ex) {
-				Console.WriteLine ("FUCK: {0}", ex);
+				Console.WriteLine ("Got a TargetException during LoadSession: {0}", ex);
 				debugger.Dispose ();
 				debugger = null;
 				throw;
 			} catch (Exception ex) {
-				Console.WriteLine ("FUCK: {0}", ex);
+				Console.WriteLine ("Got an Exception during LoadSession: {0}", ex);
 				debugger.Dispose ();
 				debugger = null;
 				throw;
@@ -418,20 +418,49 @@ namespace Mono.Debugger.Frontend
 			return true;
 		}
 
+		public bool WaitOne (Thread thread)
+		{
+			ClearInterrupt ();
+
+			WaitHandle[] handles = new WaitHandle [2];
+			handles [0] = interrupt_event;
+			handles [1] = thread.WaitHandle;
+
+			int ret = WaitHandle.WaitAny (handles);
+
+			if (ret == 0) {
+				thread.Stop ();
+				thread.WaitHandle.WaitOne ();
+				return false;
+			}
+
+			CheckLastEvent (thread);
+			return true;
+		}
+
+		public Thread WaitAll (Thread thread)
+		{
+			return WaitAll (thread, thread.GetWaitHandle (), false);
+		}
+
 		public Thread WaitAll (Thread thread, CommandResult result, bool wait)
 		{
 			if (result == null)
 				return null;
 
 			ClearInterrupt ();
+			process_event.Reset ();
 
 			Thread stopped;
 			Hashtable seen_threads = new Hashtable ();
 			Hashtable wait_list = new Hashtable ();
 
 			do {
-				foreach (Process process in Processes) {
-					foreach (Thread t in process.GetThreads ()) {
+			again:
+				Process[] processes = HasTarget ? debugger.Processes : new Process [0];
+				foreach (Process process in processes) {
+					Thread[] the_threads = process.GetThreads ();
+					foreach (Thread t in the_threads) {
 						if ((t == thread) || seen_threads.Contains (t))
 							continue;
 
@@ -445,15 +474,16 @@ namespace Mono.Debugger.Frontend
 					}
 				}
 
-				WaitHandle[] handles = new WaitHandle [wait_list.Count + 2];
+				WaitHandle[] handles = new WaitHandle [wait_list.Count + 3];
 				handles [0] = interrupt_event;
-				handles [1] = result.CompletedEvent;
+				handles [1] = process_event;
+				handles [2] = result.CompletedEvent;
 
 				Thread[] threads = new Thread [wait_list.Count];
 				wait_list.Keys.CopyTo (threads, 0);
 
 				for (int i = 0; i < wait_list.Count; i++)
-					handles [i + 2] = threads [i].WaitHandle;
+					handles [i + 3] = threads [i].WaitHandle;
 
 				int ret = WaitHandle.WaitAny (handles);
 
@@ -462,19 +492,27 @@ namespace Mono.Debugger.Frontend
 					result.Abort ();
 					result.CompletedEvent.WaitOne ();
 					break;
+				} else if (ret == 1) {
+					seen_threads = new Hashtable ();
+					wait_list = new Hashtable ();
+					process_event.Reset ();
+					goto again;
 				}
 
 				if (result.Result is Exception)
 					throw (Exception) result.Result;
 
-				if (ret == 1)
+				if (ret == 2)
 					stopped = thread;
 				else
-					stopped = threads [ret - 2];
+					stopped = threads [ret - 3];
 
 				CheckLastEvent (stopped);
 				wait_list.Remove (stopped);
-			} while (wait && (stopped != null) && (stopped != thread));
+			} while ((stopped != null) && (stopped != thread) && (wait || !stopped.IsAlive));
+
+			if (!HasTarget)
+				return stopped;
 
 			foreach (Process process in Processes) {
 				foreach (Thread t in process.GetThreads ()) {
@@ -511,7 +549,7 @@ namespace Mono.Debugger.Frontend
 
 		public RuntimeInvokeResult RuntimeInvoke (Thread thread,
 							  TargetFunctionType function,
-							  TargetClassObject object_argument,
+							  TargetStructObject object_argument,
 							  TargetObject[] param_objects,
 							  bool is_virtual, bool debug)
 		{
@@ -524,11 +562,12 @@ namespace Mono.Debugger.Frontend
 					result.Abort ();
 					return null;
 				}
-			} else {
+			} else if (debug) {
 				Wait (thread);
 
-				if (debug)
-					CheckLastEvent (thread);
+				CheckLastEvent (thread);
+			} else {
+				Wait (result);
 			}
 
 			return result;
@@ -584,23 +623,32 @@ namespace Mono.Debugger.Frontend
 
 		protected virtual void OnThreadCreated (Thread thread)
 		{
-			Print ("Process #{0} created new thread @{1}.",
-			       thread.Process.ID, thread.ID);
+			process_event.Set ();
+			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
+				Print ("Process #{0} created new thread @{1}.",
+				       thread.Process.ID, thread.ID);
 		}
 
 		protected virtual void OnThreadExited (Thread thread)
 		{
-			if (thread != thread.Process.MainThread)
+			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
 				Print ("Thread @{0} exited.", thread.ID);
 			if (thread == current_thread)
 				current_thread = null;
 		}
 
 		protected virtual void OnMainProcessCreated (Process process)
+		{
+			new ProcessEventSink (this, process);
+		}
+
+		protected virtual void OnProcessReachedMain (Process process)
 		{ }
 
 		protected virtual void OnProcessCreated (Process process)
 		{
+			process_event.Set ();
+			new ProcessEventSink (this, process);
 			Print ("Created new process #{0}.", process.ID);
 			if (current_process == null) {
 				current_process = process;
@@ -628,9 +676,9 @@ namespace Mono.Debugger.Frontend
 
 		protected virtual void OnTargetEvent (Thread thread, TargetEventArgs args)
 		{
-			if (args.Type == TargetEventType.TargetInterrupted)
+			if (args.Type == TargetEventType.TargetInterrupted) {
 				;
-			else if ((args.Type != TargetEventType.TargetStopped) || ((int) args.Data != 0))
+			} else if ((args.Type != TargetEventType.TargetStopped) || ((int) args.Data != 0))
 				Style.TargetEvent (thread, args);
 			else if ((thread.ThreadFlags & Thread.Flags.Background) != 0)
 				Style.TargetEvent (thread, args);
@@ -821,24 +869,24 @@ namespace Mono.Debugger.Frontend
 
 		public Module[] GetModules (int[] indices)
 		{
-			int pos = 0;
-			Module[] retval = new Module [indices.Length];
+			List<Module> retval = new List<Module> ();
 			Module[] modules = CurrentProcess.Modules;
 
 			foreach (int index in indices) {
-				if ((index < 0) || (index > modules.Length))
-					throw new ScriptingException ("No such module {0}.", index);
-
+				bool found = false;
 				for (int i = 0; i < modules.Length; i++) {
-					if (modules [i].ID != index)
-						continue;
-					retval [pos] = modules [i];
+					if (modules [i].ID == index) {
+						retval.Add (modules [i]);
+						found = true;
+						break;
+					}
 				}
 
-				pos ++;
+				if (!found)
+					throw new ScriptingException ("No such module {0}.", index);
 			}
 
-			return retval;
+			return retval.ToArray ();
 		}
 
 		public SourceFile[] GetSources (int[] indices)
@@ -869,6 +917,11 @@ namespace Mono.Debugger.Frontend
 			return retval;
 		}
 
+		public SourceBuffer ReadFile (string filename)
+		{
+			return source_factory.FindFile (filename);
+		}
+
 		public void ModuleOperations (Module[] modules, ModuleOperation[] operations)
 		{
 			foreach (Module module in modules) {
@@ -990,10 +1043,10 @@ namespace Mono.Debugger.Frontend
 				debugger.ThreadCreatedEvent += thread_created;
 				debugger.ThreadExitedEvent += thread_exited;
 				debugger.MainProcessCreatedEvent += main_process_created;
+				debugger.ProcessReachedMainEvent += process_reached_main;
 				debugger.ProcessCreatedEvent += process_created;
 				debugger.ProcessExitedEvent += process_exited;
 				debugger.ProcessExecdEvent += process_execd;
-				debugger.TargetOutputEvent += target_output;
 			}
 
 			public void thread_created (Debugger debugger, Thread thread)
@@ -1013,6 +1066,11 @@ namespace Mono.Debugger.Frontend
 				interpreter.OnMainProcessCreated (process);
 			}
 
+			public void process_reached_main (Debugger debugger, Process process)
+			{
+				interpreter.OnProcessReachedMain (process);
+			}
+
 			public void process_created (Debugger debugger, Process process)
 			{
 				interpreter.OnProcessCreated (process);
@@ -1032,6 +1090,17 @@ namespace Mono.Debugger.Frontend
 			{
 				interpreter.OnTargetExited ();
 			}
+		}
+
+		protected class ProcessEventSink : DebuggerMarshalByRefObject
+		{
+			Interpreter interpreter;
+
+			public ProcessEventSink (Interpreter interpreter, Process process)
+			{
+				this.interpreter = interpreter;				
+				process.TargetOutputEvent += target_output;
+			}
 
 			public void target_output (bool is_stderr, string line)
 			{
