@@ -1,11 +1,14 @@
 using System;
 using System.IO;
 using System.Collections;
+using System.Collections.Generic;
 using System.Globalization;
 using SD = System.Diagnostics;
 using ST = System.Threading;
 using System.Text;
+using System.Text.RegularExpressions;
 using NUnit.Framework;
+using NUnit.Core;
 
 using Mono.Debugger;
 using Mono.Debugger.Languages;
@@ -80,6 +83,8 @@ namespace Mono.Debugger.Tests
 			this.inferior_stdout = inferior_stdout;
 			this.inferior_stderr = inferior_stderr;
 
+			config.FollowFork = true;
+
 			queue = Queue.Synchronized (new Queue ());
 			wait_event = new ST.ManualResetEvent (false);
 
@@ -103,7 +108,7 @@ namespace Mono.Debugger.Tests
 
 		public DebuggerEvent Wait ()
 		{
-			for (int i = 0; i < 3; i++) {
+			for (int i = 0; i < 5; i++) {
 				lock (queue.SyncRoot) {
 					if (queue.Count > 0)
 						return (DebuggerEvent) queue.Dequeue ();
@@ -111,7 +116,7 @@ namespace Mono.Debugger.Tests
 					wait_event.Reset ();
 				}
 
-				wait_event.WaitOne (2500, false);
+				wait_event.WaitOne (3000, false);
 			}
 
 			return null;
@@ -177,6 +182,8 @@ namespace Mono.Debugger.Tests
 
 		public override void Print (string message)
 		{
+			if (!IsScript)
+				base.Print (message);
 		}
 	}
 
@@ -192,8 +199,21 @@ namespace Mono.Debugger.Tests
 		public readonly string ExeFileName;
 		public readonly string FileName;
 
+		public static bool Verbose = false;
+
+		Dictionary<string,int> lines;
+		Dictionary<string,int> automatic_breakpoints;
+
+		protected static readonly StreamWriter stderr;
+
 		static TestSuite ()
 		{
+			stderr = new StreamWriter (Console.OpenStandardError ());
+			stderr.AutoFlush = true;
+
+			if (Verbose)
+				stderr.WriteLine ("PID IS {0}", LibGTop.GetPid ());
+
 			Report.Initialize ();
 			Report.ReportWriter.PrintToConsole = false;
 		}
@@ -217,10 +237,74 @@ namespace Mono.Debugger.Tests
 			inferior_stderr = new LineReader ();
 		}
 
+		const string mdb_line_re = @"//\s+@MDB (LINE|BREAKPOINT):\s+(.*)$";
+
+		protected void ReadSourceFile ()
+		{
+			lines = new Dictionary<string,int> ();
+			automatic_breakpoints = new Dictionary<string,int> ();
+
+			using (StreamReader reader = new StreamReader (FileName)) {
+				string text;
+				int line = 0;
+				while ((text = reader.ReadLine ()) != null) {
+					line++;
+
+					Match match = Regex.Match (text, mdb_line_re);
+					if (!match.Success)
+						continue;
+
+					string name = match.Groups [2].Value;
+					lines.Add (name, line);
+					if (match.Groups [1].Value == "BREAKPOINT")
+						automatic_breakpoints [name] = AssertBreakpoint (line);
+				}
+			}
+		}
+
+		protected void AssertHitBreakpoint (Thread thread, string name, string function)
+		{
+			AssertHitBreakpoint (thread, automatic_breakpoints [name],
+					     function, GetLine (name));
+		}
+
+		protected void AssertStopped (Thread thread, string name, string function)
+		{
+			AssertStopped (thread, function, GetLine (name));
+		}
+
+		protected void AssertSegfault (Thread thread, string name, string function)
+		{
+			AssertSegfault (thread, function, GetLine (name));
+		}
+
+		protected int GetBreakpoint (string text)
+		{
+			return automatic_breakpoints [text];
+		}
+
+		protected int GetLine (string text)
+		{
+			int offset = 0;
+			int pos = text.IndexOf ('+');
+			if (pos > 0) {
+				offset = Int32.Parse (text.Substring (pos + 1));
+				text = text.Substring (0, pos);
+			}
+
+			if (!lines.ContainsKey (text))
+				throw new InternalError ("No such line: {0}", text);
+			return lines [text] + offset;
+		}
+
 		public NUnitInterpreter Interpreter {
 			get { return interpreter; }
 		}
 
+		public DebuggerConfiguration Config {
+			get { return config; }
+		}
+
 		public static string SourceDirectory {
 			get {
 				string srcdir = Path.Combine (BuildInfo.srcdir, "../test/src/");
@@ -258,6 +342,11 @@ namespace Mono.Debugger.Tests
 					     FileName, child.ExitCode, child.StandardError.ReadToEnd ());
 		}
 
+		public static void Debug (string message, params object[] args)
+		{
+			stderr.WriteLine (message, args);
+		}
+
 		[TestFixtureSetUp]
 		public virtual void SetUp ()
 		{
@@ -266,6 +355,8 @@ namespace Mono.Debugger.Tests
 
 			engine = interpreter.DebuggerEngine;
 			parser = new LineParser (engine);
+
+			ReadSourceFile ();
 		}
 
 		[TestFixtureTearDown]
@@ -274,6 +365,22 @@ namespace Mono.Debugger.Tests
 			interpreter.Dispose ();
 			interpreter = null;
 			GC.Collect ();
+
+			if (Verbose) {
+				string name = Path.GetFileNameWithoutExtension (FileName);
+				stderr.WriteLine ("{0}: {1}", name, PrintMemoryInfo ());
+			}
+		}
+
+		public string PrintMemoryInfo ()
+		{
+			int pid = LibGTop.GetPid ();
+			int files = LibGTop.GetOpenFiles (pid);
+			LibGTop.MemoryInfo info = LibGTop.GetMemoryInfo (pid);
+			return String.Format ("size = {0,5}  vsize = {1,7}  resident = {2,5}  " +
+					      "share = {3,5}  rss = {4,5}  files = {5,2}",
+					      info.Size, info.VirtualSize, info.Resident,
+					      info.Share, info.RSS, files);
 		}
 
 		private static DebuggerOptions CreateOptions (string application, string[] args)
@@ -339,6 +446,15 @@ namespace Mono.Debugger.Tests
 					     "but expected `{2}'.", command, exception, exp_exception);
 		}
 
+		public void AssertFrame (Thread thread, string line, string function)
+		{
+			try {
+				AssertFrame (thread.CurrentFrame, 0, function, GetLine (line));
+			} catch (TargetException ex) {
+				Assert.Fail ("Cannot get current frame: {0}.", ex.Message);
+			}
+		}
+
 		public void AssertFrame (Thread thread, string function, int line)
 		{
 			try {
@@ -354,10 +470,10 @@ namespace Mono.Debugger.Tests
 				Assert.AreEqual (level, frame.Level,
 						 "Stack frame is from level {0}, but expected {1}.",
 						 frame.Level, level);
-				if (frame.SourceAddress == null)
+				if (frame.SourceLocation == null)
 					Assert.Fail ("Current frame `{0}' has no source code.", frame);
 
-				SourceLocation location = frame.SourceAddress.Location;
+				SourceLocation location = frame.SourceLocation;
 
 				string name = location.Name;
 				int pos = name.IndexOf (':');
@@ -387,7 +503,6 @@ namespace Mono.Debugger.Tests
 
 		public void AssertTargetOutput (string line)
 		{
-			inferior_stdout.Wait ();
 			string output = inferior_stdout.ReadLine ();
 			if (output == null)
 				Assert.Fail ("No target output.");
@@ -399,9 +514,10 @@ namespace Mono.Debugger.Tests
 
 		public void AssertNoTargetOutput ()
 		{
-			string output;
-			while ((output = inferior_stdout.ReadLine ()) != null)
+			if (inferior_stdout.HasEvent) {
+				string output = inferior_stdout.ReadLine ();
 				Assert.Fail ("Got unexpected target output `{0}'.", output);
+			}
 		}
 
 		public void AssertStopped (Thread exp_thread, string exp_func, int exp_line)
@@ -417,6 +533,19 @@ namespace Mono.Debugger.Tests
 				AssertFrame (exp_thread, exp_func, exp_line);
 		}
 
+		public void AssertSegfault (Thread exp_thread, string exp_func, int exp_line)
+		{
+			TargetEventArgs args = AssertTargetEvent (
+				exp_thread, TargetEventType.TargetStopped);
+
+			if ((int) args.Data != 11)
+				Assert.Fail ("Received event {0} while waiting for {1} to stop.",
+					     args, exp_thread);
+
+			if (exp_func != null)
+				AssertFrame (exp_thread, exp_func, exp_line);
+		}
+
 		public void AssertHitBreakpoint (Thread exp_thread, int exp_index,
 						 string exp_func, int exp_line)
 		{
@@ -439,12 +568,19 @@ namespace Mono.Debugger.Tests
 
 		public int AssertBreakpoint (int location)
 		{
-			return AssertBreakpoint (location.ToString ());
+			return AssertBreakpoint (String.Format ("{0}:{1}", FileName, location));
 		}
 
 		public int AssertBreakpoint (string location)
 		{
-			object result = AssertExecute ("break " + location);
+			return AssertBreakpoint (false, location);
+		}
+
+		public int AssertBreakpoint (bool lazy, string location)
+		{
+			string command = String.Concat (
+				"break ", lazy ? "-lazy " : "", location);
+			object result = AssertExecute (command);
 			if (result == null) {
 				Assert.Fail ("Failed to insert breakpoint.");
 				return -1;
@@ -577,6 +713,8 @@ namespace Mono.Debugger.Tests
 					Assert.Fail ("Cannot resolve expression `{0}'.", expression);
 
 				return expr.EvaluateType (context);
+			} catch (ScriptingException) {
+				throw;
 			} catch (AssertionException) {
 				throw;
 			} catch (Exception ex) {
@@ -612,6 +750,28 @@ namespace Mono.Debugger.Tests
 					     expression, text, exp_result);
 		}
 
+		public void AssertPrintRegex (Thread thread, DisplayFormat format,
+					      string expression, string exp_re)
+		{
+			string text = null;
+			try {
+				ScriptingContext context = GetContext (thread);
+
+				object obj = EvaluateExpression (context, expression);
+				text = context.FormatObject (obj, format);
+			} catch (AssertionException) {
+				throw;
+			} catch (Exception ex) {
+				Assert.Fail ("Failed to print expression `{0}': {1}",
+					     expression, ex);
+			}
+
+			Match match = Regex.Match (text, exp_re);
+			if (!match.Success)
+				Assert.Fail ("Expression `{0}' evaluated to `{1}', but expected `{2}'.",
+					     expression, text, exp_re);
+		}
+
 		public void AssertPrintException (Thread thread, string expression, string exp_result)
 		{
 			string text = null;
@@ -620,7 +780,7 @@ namespace Mono.Debugger.Tests
 
 				object obj = EvaluateExpression (context, expression);
 				text = context.FormatObject (obj, DisplayFormat.Object);
-				Assert.Fail ("Evaluation of exception `{0}' failed to throw " +
+				Assert.Fail ("Evaluation of expression `{0}' failed to throw " +
 					     "exception {1}.", expression, exp_result);
 			} catch (AssertionException) {
 				throw;
@@ -656,6 +816,47 @@ namespace Mono.Debugger.Tests
 					     expression, text, exp_result);
 		}
 
+		public void AssertTypeException (Thread thread, string expression, string exp_result)
+		{
+			string text = null;
+			try {
+				ScriptingContext context = GetContext (thread);
+
+				TargetType type = EvaluateExpressionType (context, expression);
+				text = context.FormatType (type);
+				Assert.Fail ("Evaluation of expression `{0}' failed to throw " +
+					     "exception {1}.", expression, exp_result);
+			} catch (AssertionException) {
+				throw;
+			} catch (ScriptingException ex) {
+				text = ex.Message;
+			} catch (Exception ex) {
+				Assert.Fail ("Failed to print expression `{0}': {1}",
+					     expression, ex);
+			}
+
+			if (text != exp_result)
+				Assert.Fail ("Expression `{0}' evaluated to `{1}', but expected `{2}'.",
+					     expression, text, exp_result);
+		}
+
+		public object EvaluateExpression (Thread thread, string expression)
+		{
+			try {
+				ScriptingContext context = GetContext (thread);
+				object obj = EvaluateExpression (context, expression);
+				if (obj == null)
+					Assert.Fail ("Failed to evaluate expression `{0}'", expression);
+				return obj;
+			} catch (AssertionException) {
+				throw;
+			} catch (Exception ex) {
+				Assert.Fail ("Failed to print expression `{0}': {1}",
+					     expression, ex);
+				return null;
+			}
+		}
+
 		public DebuggerEvent AssertEvent ()
 		{
 			DebuggerEvent e = Interpreter.Wait ();
@@ -768,59 +969,59 @@ namespace Mono.Debugger.Tests
 
 			AssertNoTargetOutput ();
 		}
+
+		public void MainLoop ()
+		{
+			CommandLineInterpreter cmdline = new CommandLineInterpreter (interpreter);
+			cmdline.RunInferiorMainLoop ();
+		}
 	}
 
-	internal class LineReader
+	internal class LineReader : MarshalByRefObject
 	{
 		string current_line = "";
-		Queue lines = new Queue ();
+		Queue queue = Queue.Synchronized (new Queue ());
 
-		bool waiting;
 		ST.AutoResetEvent wait_event = new ST.AutoResetEvent (false);
 
 		public void Add (string text)
 		{
-			lock (this) {
+			lock (queue.SyncRoot) {
 			again:
 				int pos = text.IndexOf ('\n');
 				if (pos < 0)
 					current_line += text;
 				else {
 					current_line += text.Substring (0, pos);
-					lines.Enqueue (current_line);
+					queue.Enqueue (current_line);
 					current_line = "";
 					text = text.Substring (pos + 1);
 					if (text.Length > 0)
 						goto again;
 				}
 
-				if (!waiting)
-					return;
+				wait_event.Set ();
 			}
-
-			waiting = false;
-			wait_event.Set ();
 		}
 
-		public void Wait ()
-		{
-			lock (this) {
-				if (lines.Count > 0)
-					return;
-
-				waiting = true;
-			}
-			wait_event.WaitOne (2500, false);
+		public bool HasEvent {
+			get { return queue.Count > 0; }
 		}
 
 		public string ReadLine ()
 		{
-			lock (this) {
-				if (lines.Count < 1)
-					return null;
-				else
-					return (string) lines.Dequeue ();
+			for (int i = 0; i < 5; i++) {
+				lock (queue.SyncRoot) {
+					if (queue.Count > 0)
+						return (string) queue.Dequeue ();
+
+					wait_event.Reset ();
+				}
+
+				wait_event.WaitOne (3000, false);
 			}
+
+			return null;
 		}
 	}
 }
